#include <iostream>  // For error reporting
#include <cctype>
#include <vector>
#include <map>

#include "syntaxparser.hh"
#include "bitbuf.hh"

typedef unsigned LabelType;

static const LabelType FirstLabel = 0;

static map<string, LabelType> Variables;

static unsigned RunCommands(const string &target, unsigned ekapos=0)
{
    unsigned a=ekapos;
    for(;;)
    {
        unsigned b = target.find('=', a);
        if(b == target.npos)break;
        string varname = target.substr(a, b-a);
        
        a = b+1;

        LabelType label;
        const char *s = target.c_str() + a;
        for(label=0; isdigit(*s); ++s, ++a)label=label*10+(*s-'0');
        
        Variables[varname] = label;
        
    //  cerr << "Setting '" << varname << "' = " << label << endl;
        
        while(*s==',')++s,++a;
    }
    return a;    
}

class Choice
{
public:
    string text;
    string target;
public:
    Choice(const string &txt, const string &targ) : text(txt), target(targ)
    {
    }
public:
    inline const string &gettext() const
    {
        return text;
    }
    LabelType gettarget() const
    {
        unsigned a = RunCommands(target);
        if(isdigit(target[a]))
        {
            LabelType label;
            const char *s = target.c_str() + a;
            for(label=0; isdigit(*s); ++s)label=label*10+(*s-'0');
            return label;
        }
        return Variables[target];
    }
};

class ChoiceSet : public vector<Choice>
{
    string Commands;
public:
    void AddCommand(const string &s)
    {
        if(Commands.size())Commands += ',';
        Commands += s;
    }
    void Activate() const
    {
        RunCommands(Commands);
    }
};

class Labels : public map<LabelType, ChoiceSet>
{
};

Syntax::Syntax(const char *const s)
{
    Choices = new Labels();
    Load(s);
}
Syntax::~Syntax()
{
    delete Choices;
}

void Syntax::Load(const char *s)
{
    while(*s)
    {
        LabelType label;

        for(label=0; isdigit(*s); ++s)label=label*10+(*s-'0');

        ChoiceSet Setti;
        
        for(;;)
        {
            string text, target;
            while(*s==' '||*s=='\t')++s;
            if(*s=='#' && s[1]=='#')
            {
                for(s+=2; *s!='\n' && *s; ++s)text += *s;
                Setti.AddCommand(text);
                if(*s++ != '\n')
                {
                    cerr << "No '\\n'\n" << flush;
                    throw;
                }
            }            
            else
            {
                for(; *s!='\t'&&*s!=' '&&*s; ++s)text+=*s=='_'?' ':*s;
                while(*s==' '||*s=='\t')++s;
                for(; *s!='\n'&&*s; ++s)target += *s;
                if(*s++ != '\n')
                {
                    cerr << "No '\\n'\n" << flush;
                    throw;
                }
                Choice tmp(text, target);
                
                ChoiceSet::const_iterator i;
                for(i=Setti.begin(); i!=Setti.end(); ++i)
                {
                    if(i->gettext() == text.substr(0, i->gettext().size()))
                    {
                        cerr << "Ambiguous in " << label << ": \"" << i->gettext() << "\" and \"" << text << "\"\n" << flush;
                        throw;
                    }
                }
                
                Setti.push_back(tmp);
            }
            if(*s!=' ' && *s!='\t')break;
        }
        
        unsigned n = Setti.size();
        if((n & (n-1)) != 0)
        {
            cerr << "Entry " << label << " has non-2^ -count of choices (" << n << ")\n" << flush;
            throw;
        }
        if(Choices->find(label) != Choices->end())
        {
            cerr << "Redefinition of label " << label << endl << flush;
            throw;
        }        
        (*Choices)[label] = Setti;
    }
}

string Syntax::Encode(const BitBuf &s) const return result
{
	unsigned bitpos = 0;
    
    LabelType label = FirstLabel;
    
    while(bitpos < s.size())
    {
        Labels::const_iterator i;
        i = Choices->find(label);
        if(i == Choices->end())
        {
            cerr << "Label " << label << " not found\n" << flush;
            throw;
        }
        unsigned b=0, c = i->second.size();
        while(c > unsigned(1 << b))++b;
        unsigned n = s.Read(b, bitpos);
        bitpos += b;
        
        // cerr << "Selection count = " << c << ", fetching " << b << " bits: " << n << endl << flush;
        
        const ChoiceSet &Setti = i->second;
        
        Setti.Activate();
        
        // cerr << "Syntax::Encode(): Got '" << Setti[n].gettext() << "'\n" << flush;
        
        result += Setti[n].gettext();
        label = Setti[n].gettarget();
    }
    return result;
}

BitBuf Syntax::Decode(const string &s) const return result
{
    LabelType label = FirstLabel;

    for(unsigned srcpos=0; srcpos < s.size(); )
    {
        Labels::const_iterator i = Choices->find(label);
        if(i == Choices->end())
        {
            cerr << "Label " << label << " not found\n" << flush;
            throw;
        }
        
        i->second.Activate();
        
        unsigned a, b=0, c=i->second.size();
        while(c > unsigned(1 << b))++b;
        for(a=0; a<c; ++a)
        {
            const string &txt = i->second[a].gettext();
            unsigned txtsize = txt.size();
            if(s.substr(srcpos, txtsize) == txt)
            {
            //  cerr << "Max count = " << c << ", adding " << b << " bits: " << a << endl << flush;
                srcpos += txtsize;
                label = i->second[a].gettarget();
                break;
            }
        }
        if(a == c)
        {
            cerr << "Not found, at '" << s.substr(srcpos) << "'\n" << flush;
            throw;
        }
        result.Add(a, b);
    }
    
    return result;
}
